package mycanal

import (
	"encoding/json"
	"fmt"
	"log"
	"time"

	"github.com/go-mysql-org/go-mysql/canal"
	"github.com/go-mysql-org/go-mysql/mysql"
	"github.com/go-mysql-org/go-mysql/replication"
	"github.com/gomodule/redigo/redis"
)

type Canal struct {
	name       string
	redisPool  *redis.Pool
	rowEvCanal RowEventHandler
	c          *canal.Canal
	pos        mysql.Position
	last       bool
}
type RowEventHandler interface {
	OnRow(ev *canal.RowsEvent) error
}

func NewCanal(name, my_host, my_user, my_pass string, my_port int, redis_conn, redis_pass string, redis_db int, last bool, rowEvCanal RowEventHandler) (*Canal, error) {
	pool := &redis.Pool{
		Dial: func() (redis.Conn, error) {
			opts := []redis.DialOption{
				redis.DialConnectTimeout(time.Millisecond * 120),
				redis.DialReadTimeout(time.Millisecond * 120),
				redis.DialWriteTimeout(time.Second),
				redis.DialDatabase(redis_db),
			}
			if redis_pass != "" {
				opts = append(opts, redis.DialPassword(redis_pass))
			}
			return redis.Dial("tcp", redis_conn, opts...)
		},
		MaxIdle:     10,
		MaxActive:   1024,
		IdleTimeout: time.Minute,
	}
	cfgCanal := canal.NewDefaultConfig()
	cfgCanal.Addr = fmt.Sprintf("%s:%d", my_host, my_port)
	cfgCanal.User = my_user
	cfgCanal.Password = my_pass
	c, err := canal.NewCanal(cfgCanal)
	if err != nil {
		return nil, err
	}
	// Register a handler to handle RowsEvent
	self := &Canal{
		name:       name,
		redisPool:  pool,
		rowEvCanal: rowEvCanal,
		c:          c,
		last:       last,
	}
	conn := pool.Get()
	defer conn.Close()
	log.Println("redis name", name)
	b, err := redis.Bytes(conn.Do("HGET", "binlog_connection", name))
	if err == nil {
		json.Unmarshal(b, &self.pos)
	} else if err != redis.ErrNil {
		return nil, err
	}
	c.SetEventHandler(self)
	return self, nil
}

func (h *Canal) OnRotate(*replication.RotateEvent) error          { return nil }
func (h *Canal) OnTableChanged(schema string, table string) error { return nil }
func (h *Canal) OnDDL(nextPos mysql.Position, queryEvent *replication.QueryEvent) error {
	return nil
}
func (h *Canal) OnXID(mysql.Position) error { return nil }
func (h *Canal) OnGTID(mysql.GTIDSet) error { return nil }
func (h *Canal) OnPosSynced(pos mysql.Position, gtidset mysql.GTIDSet, force bool) error {
	if b, err := json.Marshal(pos); err == nil {
		conn := h.redisPool.Get()
		defer conn.Close()
		conn.Do("HSET", "binlog_connection", h.name, string(b))
	}
	return nil
}

func (h *Canal) OnRow(ev *canal.RowsEvent) error {
	return h.rowEvCanal.OnRow(ev)
}

func (h *Canal) String() string { return h.name }

func (h *Canal) Run() (err error) {
	if h.pos.Name != "" {
		return h.c.RunFrom(h.pos)
	}
	if h.last {
		pos, err := h.c.GetMasterPos()
		if err != nil {
			return err
		}
		if b, err := json.Marshal(pos); err == nil {
			conn := h.redisPool.Get()
			defer conn.Close()
			conn.Do("HSET", "binlog_connection", h.name, string(b))
		}
		h.pos = pos
		h.c.RunFrom(pos)
	}
	return h.c.Run()
}

func (h *Canal) Close() (err error) {
	h.c.Close()
	return nil
}
