package clickhouse

import (
	"database/sql"
	"fmt"
	"log"
	"strconv"
	"strings"
	"time"

	"github.com/jmoiron/sqlx"
)

// CKNode clickhouse节点增强功能
type CKNode struct {
	*sqlx.DB
}

// NewCKTable 初始化数据表
func (node *CKNode) NewCKTable(database, table string) *CKTable {
	ckt := &CKTable{
		database: database,
		table:    table,
		node:     node,
	}
	ckt.SetColumn()
	return ckt
}

type rowDesc struct {
	Name              string      `db:"name"`
	Type              string      `db:"type"`
	DefaultKind       string      `db:"default_kind"`
	DefaultExpression string      `db:"default_expression"`
	DefaultValue      interface{} `db:"-"`
}

// CKTable clickhouse数据表
type CKTable struct {
	insertSQL string
	desc      []rowDesc
	database  string
	table     string
	node      *CKNode
}

// SetColumn 初始化字段
func (ckt *CKTable) SetColumn() error {
	ckt.desc = ckt.desc[0:0]
	if err := ckt.node.Select(&ckt.desc, "select name, type, default_kind, default_expression from system.columns where `database`=? and `table`=?", ckt.database, ckt.table); err != nil {
		return err
	}
	for i, d := range ckt.desc {
		if d.DefaultKind != "" {
			ckt.desc[i].DefaultValue = d.DefaultExpression
			continue
		}
		switch d.Type {
		case "String":
			ckt.desc[i].DefaultValue = ""
		case "Date", "DateTime":
			ckt.desc[i].DefaultValue = time.Unix(0, 0)
		case "UInt8", "UInt16", "UInt32", "UInt64", "Int8", "Int16", "Int32", "Int64":
			ckt.desc[i].DefaultValue = 0
		case "Float32", "Float64":
			ckt.desc[i].DefaultValue = 0.0
		case "Array(String)":
			ckt.desc[i].DefaultValue = []string{}
		default:
			ckt.desc[i].DefaultValue = ""
		}
	}
	fields := make([]string, len(ckt.desc))
	questionMarks := make([]string, len(ckt.desc))
	for i := range fields {
		fields[i] = fmt.Sprintf("`%s`", ckt.desc[i].Name)
		questionMarks[i] = "?"
	}
	ckt.insertSQL = fmt.Sprintf("INSERT INTO %s.%s (%s) VALUES (%s)", ckt.database, ckt.table, strings.Join(fields, ","), strings.Join(questionMarks, ","))
	log.Println("query: ", ckt.insertSQL)
	return nil
}

func (ckt *CKTable) setArgs(ev map[string]interface{}) []interface{} {
	args := make([]interface{}, len(ckt.desc))
	for i, field := range ckt.desc {
		if v, ok := ev[field.Name]; ok && v != nil {
			switch field.Type {
			case "String", "Date", "DateTime", "Array(String)":
				args[i] = fmt.Sprintf("%v", v)
			case "UInt8", "UInt16", "UInt32", "UInt64", "Int8", "Int16", "Int32", "Int64":
				if f, ok := v.(float64); ok {
					args[i] = int64(f)
				} else if s, ok := v.(string); ok {
					args[i], _ = strconv.ParseInt(s, 10, 64)
				} else {
					args[i] = v
				}
			default:
				args[i] = v
			}
		} else {
			args[i] = field.DefaultValue
		}
	}
	return args
}

func (ckt *CKTable) doBatchInsert(events []map[string]interface{}) error {
	var tx *sql.Tx
	var err error
	if tx, err = ckt.node.Begin(); err != nil {
		return err
	}

	var stmt *sql.Stmt
	if stmt, err = tx.Prepare(ckt.insertSQL); err != nil {
		tx.Rollback()
		return err
	}
	defer stmt.Close()

	for _, event := range events {
		if _, err = stmt.Exec(ckt.setArgs(event)...); err != nil {
			tx.Rollback()
			return err
		}
	}

	return tx.Commit()
}

// BatchInsert 批量插入
func (ckt *CKTable) BatchInsert(batchSize int, events []map[string]interface{}) ([]map[string]interface{}, error) {
	for {
		if len(events) > batchSize {
			if err := ckt.doBatchInsert(events[0:batchSize]); err != nil {
				return events, err
			}
			events = events[batchSize:]
		} else {
			if err := ckt.doBatchInsert(events); err != nil {
				return events, err
			}
			return nil, nil
		}
	}
}
