package clickhouse

import (
	"database/sql"
	"fmt"
	"sync"
	"time"

	"github.com/jmoiron/sqlx"

	// clichouse
	_ "github.com/ClickHouse/clickhouse-go"
)

type ckNode struct {
	Host    string `db:"host_address"`
	Port    int    `db:"port"`
	IsLocal int    `db:"is_local"`
}

// ClickHouseCluster clickhouse集群
type ClickHouseCluster struct {
	Name        string
	dbs         map[string]*sqlx.DB
	hosts       []string
	masters     []*sqlx.DB
	robinIdx    int
	robinLocker sync.Mutex
}

// NewCluster 创建集群句柄
func NewCluster(name string, addrs []string, user, pass string) (*ClickHouseCluster, error) {
	cluster := &ClickHouseCluster{Name: name, dbs: make(map[string]*sqlx.DB)}
	var errs []error
	for _, addr := range addrs {
		host := addr[len("tcp://"):]
		if db, err := sqlx.Connect("clickhouse", fmt.Sprintf("%s?username=%s&password=%s", addr, user, pass)); err == nil {
			cluster.masters = append(cluster.masters, db)
			cluster.dbs[host] = db
			cluster.hosts = append(cluster.hosts, host)
		} else {
			errs = append(errs, err)
		}
	}

	if len(cluster.masters) == 0 {
		return nil, fmt.Errorf("no valid master: %s", errs[0].Error())
	}
	if err := cluster.setCluster(user, pass); err != nil {
		return nil, err
	}
	if len(cluster.dbs) == 0 {
		return nil, fmt.Errorf("no valid node")
	}
	return cluster, nil
}

func (ck *ClickHouseCluster) setCluster(user, pass string) error {
	for _, db := range ck.masters {
		var ckNodes []*ckNode
		if err := db.Select(&ckNodes, "select host_address, port,is_local from system.clusters where cluster=?", ck.Name); err == nil {
			if len(ckNodes) == len(ck.dbs) {
				return nil
			}
			for _, node := range ckNodes {
				if node.IsLocal == 0 {
					if _, ok := ck.dbs[node.Host]; ok {
						continue
					}
					ndb, err := sqlx.Connect("clickhouse", fmt.Sprintf("tcp://%s:%d?username=%s&password=%s", node.Host, node.Port, user, pass))
					if err != nil {
						return err
					}
					ck.dbs[node.Host] = ndb
					ck.hosts = append(ck.hosts, node.Host)
				}
			}
			return nil
		}
	}
	return fmt.Errorf("query node all fail: %d", len(ck.masters))
}

// ExecOnLeader leader节点执行语句
func (ck *ClickHouseCluster) ExecOnLeader(database, table, query string, args ...interface{}) error {
	for _, db := range ck.dbs {
		var isLeader int
		if err := db.Get(&isLeader, "select is_leader from system.replicas where database=? and table=?", database, table); err != nil {
			return err
		}
		println("ExecOnLeader")
		if _, err := db.Exec(query, args...); err != nil {
			return err
		}
	}
	return nil
}

func (ck *ClickHouseCluster) Select(dest interface{}, query string, args ...interface{}) (err error) {
	for _, db := range ck.masters {
		err = db.Select(dest, query, args...)
		if err == nil {
			return nil
		}
	}
	return err
}

func (ck *ClickHouseCluster) Get(dest interface{}, query string, args ...interface{}) (err error) {
	for _, db := range ck.masters {
		err = db.Get(dest, query, args...)
		if err == nil {
			return nil
		}
	}
	return err
}

// ClickHouseClusterFunc 执行函数
type ClickHouseClusterFunc func(db *sqlx.DB) error

// ExecFunc 在所有节点上执行任务
func (ck *ClickHouseCluster) ExecFunc(ckfunc ClickHouseClusterFunc) error {
	for _, db := range ck.dbs {
		if err := ckfunc(db); err != nil {
			return err
		}
	}
	return nil
}

// RobinDB round robin 方式获取一个节点
func (ck *ClickHouseCluster) RobinDB() (db *sqlx.DB) {
	ck.robinLocker.Lock()
	defer ck.robinLocker.Unlock()
	ck.robinIdx++
	ck.robinIdx = ck.robinIdx % len(ck.dbs)
	host := ck.hosts[ck.robinIdx]
	db = ck.dbs[host]
	return
}

// ExecFuncOnRobinNode 在Round robin 一个节点上执行
func (ck *ClickHouseCluster) ExecFuncOnRobinNode(ckfunc ClickHouseClusterFunc) (err error) {
	for i := 0; i < len(ck.dbs); i++ {
		db := ck.RobinDB()
		err = ckfunc(db)
		if err == nil {
			break
		}
	}
	return err
}

// RobinGet 在Round robin 一个节点上执行Get
func (ck *ClickHouseCluster) RobinGet(dest interface{}, query string, args ...interface{}) (err error) {
	for i := 0; i < len(ck.dbs); i++ {
		db := ck.RobinDB()
		err = db.Get(dest, query, args...)
		if err == nil {
			break
		}
	}
	return err
}

// RobinSelect 在Round robin 一个节点上执行Select
func (ck *ClickHouseCluster) RobinSelect(dest interface{}, query string, args ...interface{}) (err error) {
	for i := 0; i < len(ck.dbs); i++ {
		db := ck.RobinDB()
		err = db.Select(dest, query, args...)
		if err == nil {
			break
		}
	}
	return err
}

// ExecFuncOnLeader 在leader节点上执行任务
func (ck *ClickHouseCluster) ExecFuncOnLeader(database, table string, ckfunc ClickHouseClusterFunc) error {
	for _, db := range ck.dbs {
		var isLeader int
		if err := db.Get(&isLeader, "select is_leader from system.replicas where database=? and table=?", database, table); err != nil && err != sql.ErrNoRows {
			return err
		}
		if isLeader == 1 {
			if err := ckfunc(db); err != nil {
				return err
			}
		}
	}
	return nil
}

func (ck *ClickHouseCluster) ExecFuncOnMaster(ckfunc ClickHouseClusterFunc) (err error) {
	for _, master := range ck.masters {
		err = ckfunc(master)
		if err == nil {
			return
		}
	}
	return
}

// CheckMutation 检查mutation数量
func (ck *ClickHouseCluster) CheckMutation(dbname, table string) (int, error) {
	for _, db := range ck.dbs {
		var count int
		if err := db.Get(&count, "select count() from system.mutations where database=? and table=? and is_done=0", dbname, table); err != nil && err != sql.ErrNoRows {
			return 0, err
		}
		if count > 0 {
			return count, nil
		}
	}
	return 0, nil
}

// ErrMutationDuplicate mutation未结束
var ErrMutationDuplicate = fmt.Errorf("mutation duplicate")

// WaitMutation 等待Mutation结束
func (ck *ClickHouseCluster) WaitMutation(dbname, table string, retry int) error {
	for i := 0; i < retry; i++ {
		time.Sleep(time.Second)
		count, err := ck.CheckMutation(dbname, table)
		if err != nil {
			return err
		}
		if count == 0 {
			return nil
		}
	}
	return ErrMutationDuplicate
}

// KillMutation 杀死mutation进程
func (ck *ClickHouseCluster) KillMutation(dbname, table string) error {
	_, err := ck.masters[0].Exec(fmt.Sprintf("kill mutation on cluster %s where database=? and table=? sync", ck.Name), dbname, table)
	return err
}
