package exception

import (
	"bytes"
	"context"
	"encoding/json"
	"fmt"
	"html/template"
	"io/ioutil"
	"log"
	"net/http"
	"strconv"
	"time"

	"github.com/getsentry/sentry-go"
	"github.com/labstack/echo/v4"
	"github.com/spf13/viper"
	cmq "github.com/yougg/cmq-go-tdmq"
	"gitlab-ce.k8s.tools.vchangyi.com/cdp/cy-sdk-go/crypt"
	"gitlab-ce.k8s.tools.vchangyi.com/cdp/cy-sdk-go/wechat"
	"gopkg.in/gomail.v2"
)

// AlertContent 邮件内容
type AlertContent struct {
	Topic   string        `json:"topc"`
	Host    string        `json:"host"`
	PID     int           `json:"pid"`
	Content string        `json:"content"`
	Time    string        `json:"time"`
	Args    []interface{} `json:"args,omitempty"`
}

type cmdQueue struct {
	client *cmq.Client
	queue  string
}

var defaultTemplate = template.Must(template.New("default").Parse("主题: {{.Topic}} 发生报警 {{.Contents | len}} 次\n{{range .Contents}}报警时间:\t{{.Time}}\n主机名称:\t{{.Host}}\n进程ID:\t{{.PID}}\n报警内容:\t{{.Content}}\n{{range .Args}}\t-\t{{.}}\n {{end}} {{end}}"))

// Hash 生成过滤hash
func (ac AlertContent) Hash(hashAlgo string) string {
	switch hashAlgo {
	case "topic":
		return crypt.MD5Hash(ac.Topic + ac.Host)
	case "process":
		return crypt.MD5Hash(ac.Topic + ac.Host + strconv.Itoa(ac.PID))
	case "content":
		return crypt.MD5Hash(ac.Topic + ac.Host + ac.Content)
	case "arg0":
		return crypt.MD5Hash(fmt.Sprintf("%s%s%s%v", ac.Topic, ac.Host, ac.Content, ac.Args[0]))
	}
	return ""
}

var alertChan = make(chan *AlertContent, 64)
var alertQueue *cmdQueue

func sendContents(queue string, contents []*AlertContent) {
	if alertQueue == nil {
		b, err := json.Marshal(contents)
		if err != nil {
			sentry.CaptureException(err)
		} else {
			sentry.CaptureMessage(string(b))
		}
	} else {
		var msgs []string
		for _, content := range contents {
			b, _ := json.Marshal(content)
			msgs = append(msgs, string(b))
		}
		rsp, err := alertQueue.client.BatchSendMessage(queue, msgs, 0)
		log.Println(rsp, err)
	}
}

// SetAlertConfig 显式设置消息队列
func SetAlertConfig(queue string, endpoint string, secretID string, secretKey string) {
	alertQueue = &cmdQueue{
		client: cmq.NewClient(endpoint, secretID, secretKey, 5*time.Second),
		queue:  queue,
	}
	var contents []*AlertContent
	go func() {
		for {
			select {
			case content := <-alertChan:
				contents = append(contents, content)
				if len(contents) >= 16 {
					go sendContents(queue, contents)
					contents = contents[0:0]
				}
			case <-time.After(time.Second * 5):
				if len(contents) > 0 {
					go sendContents(queue, contents)
					contents = contents[0:0]
				}
			}
		}

	}()
}

// SetAlertViper 自动获取viper配置
func SetAlertViper() {
	SetAlertConfig(viper.GetString("alert_queue"), viper.GetString("alert_endpoint"), viper.GetString("alert_secret_id"), viper.GetString("alert_secret_key"))
}

// SendAlert 发送报警内容
func SendAlert(topic string, err error, args ...interface{}) {
	SendContent(topic, err.Error(), args...)
}

// SendContent 发送内容
func SendContent(topic string, content string, args ...interface{}) {
	alertChan <- &AlertContent{
		Topic:   topic,
		Content: content,
		Host:    host,
		PID:     pid,
		Time:    time.Now().Format(time.RFC3339),
		Args:    args,
	}
}

func receiveAlert() ([]cmq.Message, bool) {
	rsp, err := alertQueue.client.BatchReceiveMessage(alertQueue.queue, 10, 1)
	if err == context.DeadlineExceeded {
		return nil, true
	}
	if err != nil {
		return nil, true
	}
	return rsp.MsgInfos(), len(rsp.MsgInfos()) < 10
}

// alertTopic 告警主题配置
type alertTopic struct {
	Mails    []string `yaml:"mails"`
	Robots   []string `yaml:"robots"`
	Filter   string   `yaml:"filter"`
	Template string   `yaml:"template"`
	tpl      *template.Template
}

// 异常监听
func HttpServer() {
	address := viper.GetString("http_server")
	if len(address) > 0 {
		SetAlertViper()
		e := echo.New()
		e.Use(func(next echo.HandlerFunc) echo.HandlerFunc {
			return func(c echo.Context) (err error) {
				req := c.Request()
				res := c.Response()
				reqBody := []byte{}
				defer func() {
					fmt.Printf(`{"timestamp":%d,"method":"%s","url":"%s","status":%d,"body":%s}\n`, time.Now().Unix(), req.Method, req.URL.Path, res.Status, string(reqBody))
				}()
				if c.Request().Body != nil { // Read
					reqBody, _ = ioutil.ReadAll(c.Request().Body)
				}
				c.Request().Body = ioutil.NopCloser(bytes.NewBuffer(reqBody)) // Reset
				if err = next(c); err != nil {
					c.Error(err)
				}
				return
			}
		})
		e.POST("/exception/report", func(c echo.Context) error {
			req := struct {
				Topic   string `json:"topic"`
				Message string `json:"message"`
				Host    string `json:"host,omitempty"`
				PID     int    `json:"pid,omitempty"`
			}{}
			c.Bind(&req)
			if len(req.Topic) == 0 || len(req.Message) == 0 {
				return c.JSON(http.StatusBadRequest, "error: topic和message 不能为空")
			}
			alertChan <- &AlertContent{
				Topic:   req.Topic,
				Content: req.Message,
				Host:    req.Host,
				PID:     req.PID,
				Time:    time.Now().Format(time.RFC3339),
			}
			return c.JSON(http.StatusOK, "ok")
		})
		if err := e.Start(address); err != nil {
			log.Println("start http server on ", address, "error", err)
			return
		}
		log.Println("start http server on ", address)
	}

}

// ProcessAlert 处理异常
func ProcessAlert() {
	var alertTopicMap map[string]*alertTopic
	Assert(viper.UnmarshalKey("alert_topics", &alertTopicMap))
	for topic, at := range alertTopicMap {
		if at.Template != "" {
			at.tpl = template.Must(template.New(topic).Parse(at.Template))
		} else {
			at.tpl = defaultTemplate
		}
	}

	alertQueue = &cmdQueue{
		client: cmq.NewClient(viper.GetString("alert_endpoint"), viper.GetString("alert_secret_id"), viper.GetString("alert_secret_key"), 5*time.Second),
		queue:  viper.GetString("alert_queue"),
	}
	mailConn := gomail.NewPlainDialer(viper.GetString("alert_mail_host"), viper.GetInt("alert_mail_port"), viper.GetString("alert_mail_user"), viper.GetString("alert_mail_pass"))
	hashMap := make(map[string]int64)
	for {
		var candidate []cmq.Message
		for {
			msgs, send := receiveAlert()
			if len(msgs) > 0 {
				candidate = append(candidate, msgs...)
			}
			if send || len(candidate) >= 1024 {
				break
			}
		}
		if len(candidate) == 0 {
			time.Sleep(time.Minute)
			continue
		}

		var contents = make(map[string][]*AlertContent)
		var receipts []string
		for _, msg := range candidate {
			log.Println(msg.MsgId(), msg.Handle(), msg.MsgBody())
			receipts = append(receipts, msg.Handle())
			content := &AlertContent{}
			if err := json.Unmarshal([]byte(msg.MsgBody()), content); err != nil {
				log.Println(msg.MsgBody(), "unmarshal error", err)
				sentry.CaptureException(err)
				continue
			}
			contents[content.Topic] = append(contents[content.Topic], content)
		}

		ts := time.Now().Unix()
		for topic, topicContents := range contents {
			if alertTp, ok := alertTopicMap[topic]; ok {
				hasMessage := false
				for _, content := range topicContents {
					if hashVal := content.Hash(alertTp.Filter); hashVal != "" {
						if hashMap[hashVal] > ts-1800 {
							continue
						} else {
							hashMap[hashVal] = ts
							hasMessage = true
						}
					} else {
						hasMessage = true
					}
				}
				if !hasMessage {
					log.Println("topic:", topic, "has no message")
					continue
				}
				msg := bytes.NewBuffer(nil)
				alertTp.tpl.Execute(msg, struct {
					Topic    string
					Contents []*AlertContent
				}{Topic: topic, Contents: topicContents})
				if len(alertTp.Mails) > 0 {
					m := gomail.NewMessage()
					m.SetHeader("From", viper.GetString("alert_mail_user"))
					m.SetHeader("To", alertTp.Mails...)
					m.SetHeader("Subject", fmt.Sprintf("服务器报警: %s", topic))
					m.SetBody("text/plain", msg.String())
					err := mailConn.DialAndSend(m)
					log.Println("send mail", alertTp.Mails, msg.String(), err)
					if err != nil {
						sentry.CaptureException(err)
					}
				}
				if len(alertTp.Robots) > 0 {
					for _, robot := range alertTp.Robots {
						err := wechat.SendQYWechatRobotMarkDown(robot, msg.String())
						log.Println("send robot", robot, msg.String(), err)
						if err != nil {
							sentry.CaptureException(err)
						}
					}
				}
			} else {
				log.Println("can't find topic:", topic)
			}
		}
		for i := 0; i < len(receipts); i += 16 {
			j := i + 16
			if j > len(receipts) {
				j = len(receipts)
			}
			rsp, err := alertQueue.client.BatchDeleteMessage(alertQueue.queue, receipts[i:j])
			if err == nil && rsp.Code() != 0 {
				err = fmt.Errorf("requestId(%s),response(%v)", rsp.RequestId(), rsp.Code())
			}
			if err != nil && err != context.DeadlineExceeded {
				log.Println("delete receipts ", receipts[i:j], err)
			}
		}

	}
}
