package logger

import (
	"bufio"
	"bytes"
	"encoding/json"
	"io"
	"io/ioutil"
	"net"
	"net/http"
	"strconv"
	"time"

	"github.com/labstack/echo/v4"
	"github.com/labstack/echo/v4/middleware"
)

// EchoLoggerConfig defines the config for Logger middleware.
type EchoLoggerConfig struct {
	// Skipper defines a function to skip middleware.
	Skipper middleware.Skipper
	Output  io.Writer

	// WithBody 是否记录消息内容
	WithBody bool

	// Header 消息头
	Header map[string]string
}

type bodyDumpResponseWriter struct {
	io.Writer
	http.ResponseWriter
}

func (w *bodyDumpResponseWriter) WriteHeader(code int) {
	w.ResponseWriter.WriteHeader(code)
}

func (w *bodyDumpResponseWriter) Write(b []byte) (int, error) {
	return w.Writer.Write(b)
}

func (w *bodyDumpResponseWriter) Flush() {
	w.ResponseWriter.(http.Flusher).Flush()
}

func (w *bodyDumpResponseWriter) Hijack() (net.Conn, *bufio.ReadWriter, error) {
	return w.ResponseWriter.(http.Hijacker).Hijack()
}

// EchoLogger returns a middleware that logs HTTP requests.
func EchoLogger() echo.MiddlewareFunc {
	return EchoLoggerWithConfig(EchoLoggerConfig{})
}

type logEvent struct {
	Ts       int64             `json:"ts"`
	IP       string            `json:"ip"`
	Protocol string            `json:"protocol"`
	Method   string            `json:"method"`
	Host     string            `json:"host"`
	Path     string            `json:"path"`
	IB       int               `json:"ib"`
	OB       int64             `json:"ob"`
	Req      json.RawMessage   `json:"req,omitempty"`
	Resp     json.RawMessage   `json:"resp,omitempty"`
	Status   int               `json:"status"`
	Latency  int               `json:"latency"`
	User     interface{}       `json:"user,omitempty"`
	Option   interface{}       `json:"option,omitempty"`
	Query    map[string]string `json:"query,omitempty"`
	Header   map[string]string `json:"header,omitempty"`
	Error    error             `json:"err,omitempty"`
}

// EchoLoggerWithConfig returns a Logger middleware with config.
// See: `Logger()`.
func EchoLoggerWithConfig(config EchoLoggerConfig) echo.MiddlewareFunc {
	// Defaults
	if config.Skipper == nil {
		config.Skipper = middleware.DefaultSkipper
	}
	if config.Output == nil {
		config.Output = NewRotateWriter("echo", RotateDay)
	}

	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) (err error) {
			if config.Skipper(c) {
				return next(c)
			}

			req := c.Request()
			res := c.Response()
			start := time.Now()

			ev := &logEvent{}
			if cl := req.Header.Get(echo.HeaderContentLength); cl != "" {
				ev.IB, _ = strconv.Atoi(cl)
			}

			ev.Ts = start.Unix()
			ev.Protocol = req.Proto
			ev.Host = req.Host
			ev.Method = req.Method
			ev.Path = req.URL.Path
			ev.IP = c.RealIP()

			defer func() {
				b, _ := json.Marshal(ev)
				config.Output.Write(b)
				config.Output.Write([]byte("\n"))
			}()

			if config.WithBody {
				reqBody := []byte{}
				if c.Request().Body != nil { // Read
					reqBody, _ = ioutil.ReadAll(c.Request().Body)
					c.Request().Body.Close()
				}
				c.Request().Body = ioutil.NopCloser(bytes.NewBuffer(reqBody)) // Reset
				ev.Req = reqBody

				// Response
				resBody := new(bytes.Buffer)
				mw := io.MultiWriter(res.Writer, resBody)
				res.Writer = &bodyDumpResponseWriter{Writer: mw, ResponseWriter: res.Writer}
				defer func() { ev.Resp = resBody.Bytes() }()
			}

			query := c.QueryParams()
			if len(query) > 0 {
				ev.Query = make(map[string]string)
				for k := range query {
					ev.Query[k] = query.Get(k)
				}
			}

			if len(config.Header) > 0 {
				ev.Header = make(map[string]string)
				for k, flattern := range config.Header {
					if v := req.Header.Get(k); v != "" {
						ev.Header[flattern] = v
					}
				}
			}

			if err = next(c); err != nil {
				ev.Error = err
				c.Error(err)
			}
			ev.User = c.Get("user")
			ev.Option = c.Get("option")
			ev.Status = res.Status
			ev.Latency = int(time.Now().Sub(start) / time.Millisecond)
			ev.OB = res.Size

			return
		}
	}
}
