package region

import (
	"fmt"
	"log"
	"time"

	"github.com/jmoiron/sqlx"
)

const (
	LevelCountry  = 0
	LevelProvince = 1
	LevelCity     = 2
	LevelDistrict = 3
)

type Region struct {
	ID                 int                `db:"id"`
	Name               string             `db:"sname"`
	Pinyin             string             `db:"pinyin"`
	Pid                int                `db:"pid"`
	Level              int                `db:"level"`
	subRegionNameMap   map[string]*Region `db:"-"`
	subRegionPinyinMap map[string]*Region `db:"-"`
}

func NewRegion(db *sqlx.DB) (region *Region, err error) {
	region = &Region{
		subRegionNameMap:   make(map[string]*Region, 0),
		subRegionPinyinMap: make(map[string]*Region, 0),
	}
	err = region.init(db)
	return region, err
}

func (r *Region) init(db *sqlx.DB) (err error) {
	now := time.Now()
	defer func() {
		log.Println("region init use time:", time.Now().Sub(now))
	}()
	regions := make([]*Region, 0)
	pmap := make(map[int][]*Region, 0)
	err = db.Select(&regions, "select id,pid,sname,level,pinyin from region")
	if err != nil {
		return err
	}
	for _, r := range regions {
		r.subRegionNameMap = make(map[string]*Region, 0)
		r.subRegionPinyinMap = make(map[string]*Region, 0)
		if _, ok := pmap[r.Pid]; !ok {
			pmap[r.Pid] = make([]*Region, 0)
		}
		pmap[r.Pid] = append(pmap[r.Pid], r)
	}
	var addsubRegion func(*Region) error
	addsubRegion = func(base *Region) error {
		if subrs, ok := pmap[base.ID]; ok {
			for _, subr := range subrs {
				if err := addsubRegion(subr); err != nil {
					return err
				}
				if err := base.AddsubRegion(subr); err != nil {
					return err
				}
			}
		}
		return nil
	}
	return addsubRegion(r)
}

func (a *Region) AddsubRegion(subRegion *Region) error {
	if len(subRegion.Name) == 0 || len(subRegion.Pinyin) == 0 {
		log.Println(fmt.Sprintf("Aare 信息不完整：(id:%d,name:%s,pinyin:%s)", subRegion.ID, subRegion.Name, subRegion.Pinyin))
		return nil
	}
	a.subRegionNameMap[subRegion.Name] = subRegion
	a.subRegionPinyinMap[subRegion.Pinyin] = subRegion
	return nil
}

func (a *Region) GetSubRegionByName(name string) *Region {
	return a.subRegionNameMap[name]
}
func (a *Region) GetSubRegionByPinyin(pinyin string) *Region {
	return a.subRegionPinyinMap[pinyin]
}

func (a *Region) RangeSub(fun func(sub *Region) bool) {
	for _, sub := range a.subRegionNameMap {
		if !fun(sub) {
			break
		}
	}
}
