package schedule

import (
	"reflect"
	"time"

	gocron "github.com/robfig/cron"
)

type TimeTaskOption func(c *TimerTask)
type ErrHandler func(name string, err error)
type TaskStatusHandler func(name string, nextAt time.Time)

type ErrHandlerWithContext func(context TaskContext, err error)
type TaskStatusHandlerWithContext func(context TaskContext)

type TimeTaskFunc func(...interface{}) error

type TaskContext struct {
	id       int64
	Name     string
	fuc      TimeTaskFunc
	args     []interface{}
	Duration time.Duration
}

func WithErrHandler(errHandler ErrHandler) TimeTaskOption {
	return func(tt *TimerTask) {
		tt.errHandler = errHandler
	}
}
func WithRunHandler(runHandler TaskStatusHandler) TimeTaskOption {
	return func(tt *TimerTask) {
		tt.runHandler = runHandler
	}
}

func WithEndHandler(runHandler TaskStatusHandler) TimeTaskOption {
	return func(tt *TimerTask) {
		tt.endHandler = runHandler
	}
}

func WithContextErrHandler(errHandler ErrHandlerWithContext) TimeTaskOption {
	return func(tt *TimerTask) {
		tt.errContextHandler = errHandler
	}
}
func WithContextRunHandler(runHandler TaskStatusHandlerWithContext) TimeTaskOption {
	return func(tt *TimerTask) {
		tt.runContextHandler = runHandler
	}
}

func WithContextEndHandler(runHandler TaskStatusHandlerWithContext) TimeTaskOption {
	return func(tt *TimerTask) {
		tt.endContextHandler = runHandler
	}
}

type TimerTask struct {
	tws               []*TaskContext
	cases             []reflect.SelectCase
	durations         map[*TaskContext]time.Duration
	errHandler        ErrHandler
	runHandler        TaskStatusHandler
	endHandler        TaskStatusHandler
	errContextHandler ErrHandlerWithContext
	runContextHandler TaskStatusHandlerWithContext
	endContextHandler TaskStatusHandlerWithContext
	cron              *gocron.Cron
}

// type taskWrapper struct {
// 	id   int64
// 	fuc  func() error
// 	name string
// }

func NewTimerTask(options ...TimeTaskOption) *TimerTask {
	tt := &TimerTask{
		tws:       make([]*TaskContext, 0, 0),
		cases:     make([]reflect.SelectCase, 0, 0),
		durations: make(map[*TaskContext]time.Duration, 0),
	}
	for _, opt := range options {
		opt(tt)
	}
	tt.cron = gocron.New()
	return tt
}

/*
firstAt :首次执行时间,<=当前时间，立即执行
duration: 之后执行的时间间隔
task: 任务
taskName: 任务名称
*/
func (tt *TimerTask) Do(firstAt time.Time, duration time.Duration, task TimeTaskFunc, taskName string, args ...interface{}) error {
	now := time.Now()
	at := now
	if firstAt.After(now) {
		at = firstAt
	}
	t := time.NewTimer(at.Sub(now))
	tw := &TaskContext{id: at.Unix(), fuc: task, Name: taskName, args: args}
	tt.cases = append(tt.cases, reflect.SelectCase{Dir: reflect.SelectRecv, Chan: reflect.ValueOf(t.C)})
	tt.tws = append(tt.tws, tw)
	tt.durations[tw] = duration
	return nil
}

/*
表达式说明：
Usage
Callers may register Funcs to be invoked on a given schedule.  Cron will run
them in their own goroutines.
Cron("0 30 * * * *", func(arg ...interface{}) { fmt.Println("Every hour on the half hour") }("args..."))
Cron("@hourly",      func(arg ...interface{}) { fmt.Println("Every hour") }("args..."))
Cron("@every 1h30m", func(arg ...interface{}) { fmt.Println("Every hour thirty")("args...") })
// Funcs are invoked in their own goroutine, asynchronously.
...
// Funcs may also be added to a running Cron
Cron("@daily",  func(arg ...interface{}) { fmt.Println("Every day") }("args..."))

CRON Expression Format

A cron expression represents a set of times, using 6 space-separated fields.

Field name   | Mandatory? | Allowed values  | Allowed special characters
----------   | ---------- | --------------  | --------------------------
Seconds      | Yes        | 0-59            | * / , -
Minutes      | Yes        | 0-59            | * / , -
Hours        | Yes        | 0-23            | * / , -
Day of month | Yes        | 1-31            | * / , - ?
Month        | Yes        | 1-12 or JAN-DEC | * / , -
Day of week  | Yes        | 0-6 or SUN-SAT  | * / , - ?

Note: Month and Day-of-week field values are case insensitive.  "SUN", "Sun",

and "sun" are equally accepted.
Special Characters
Asterisk ( * )
The asterisk indicates that the cron expression will match for all values of the
field; e.g., using an asterisk in the 5th field (month) would indicate every
month.
Slash ( / )
Slashes are used to describe increments of ranges. For example 3-59/15 in the
1st field (minutes) would indicate the 3rd minute of the hour and every 15
minutes thereafter. The form "*\/..." is equivalent to the form "first-last/...",
that is, an increment over the largest possible range of the field.  The form
"N/..." is accepted as meaning "N-MAX/...", that is, starting at N, use the
increment until the end of that specific range.  It does not wrap around.
Comma ( , )
Commas are used to separate items of a list. For example, using "MON,WED,FRI" in
the 5th field (day of week) would mean Mondays, Wednesdays and Fridays.
Hyphen ( - )
Hyphens are used to define ranges. For example, 9-17 would indicate every
hour between 9am and 5pm inclusive.
Question mark ( ? )
Question mark may be used instead of '*' for leaving either day-of-month or
day-of-week blank.
Predefined schedules

You may use one of several pre-defined schedules in place of a cron expression.
Entry                  | Description                                | Equivalent To
-----                  | -----------                                | -------------
@yearly (or @annually) | Run once a year, midnight, Jan. 1st        | 0 0 0 1 1 *
@monthly               | Run once a month, midnight, first of month | 0 0 0 1 * *
@weekly                | Run once a week, midnight on Sunday        | 0 0 0 * * 0
@daily (or @midnight)  | Run once a day, midnight                   | 0 0 0 * * *
@hourly                | Run once an hour, beginning of hour        | 0 0 * * * *


Intervals
You may also schedule a job to execute at fixed intervals, starting at the time it's added
or cron is run. This is supported by formatting the cron spec like this:

*******************************************************************************************
@every <duration>

where "duration" is a string accepted by time.ParseDuration
(http://golang.org/pkg/time/#ParseDuration).
For example, "@every 1h30m10s" would indicate a schedule that activates immediately,
and then every 1 hour, 30 minutes, 10 seconds.

Note: The interval does not take the job runtime into account.  For example,
if a job takes 3 minutes to run, and it is scheduled to run every 5 minutes,
it will have only 2 minutes of idle time between each run.

Time zones
All interpretation and scheduling is done in the machine's local time zone (as
provided by the Go time package (http://www.golang.org/pkg/time).
Be aware that jobs scheduled during daylight-savings leap-ahead transitions will
not be run!
Thread safety
Since the Cron service runs concurrently with the calling code, some amount of
care must be taken to ensure proper synchronization.
All cron methods are designed to be correctly synchronized as long as the caller
ensures that invocations have a clear happens-before ordering between them.
Implementation
Cron entries are stored in an array, sorted by their next activation time.  Cron
sleeps until the next job is due to be run.
Upon waking:
- it runs each entry that is active on that second
- it calculates the next run times for the jobs that were run
- it re-sorts the array of entries by next activation time.
- it goes to sleep until the soonest job.
*/
func (tt *TimerTask) Cron(spec string, task TimeTaskFunc, taskName string, args ...interface{}) error {
	ctx := TaskContext{id: time.Now().Unix(), fuc: task, Name: taskName, args: args}
	return tt.cron.AddFunc(spec, func() {
		sch, _ := gocron.Parse(spec)
		next := sch.Next(time.Now())
		if tt.runHandler != nil {
			tt.runHandler(taskName, next)
		}
		if tt.runContextHandler != nil {
			tt.runContextHandler(ctx)
		}
		err := task(args)
		if tt.endHandler != nil {
			tt.endHandler(taskName, next)
		}
		if tt.endContextHandler != nil {
			tt.endContextHandler(ctx)
		}
		if err != nil {
			if tt.errHandler != nil {
				tt.errHandler(taskName, err)
			}
			if tt.errContextHandler != nil {
				tt.errContextHandler(ctx, err)
			}
		}
	})
}

func (tt *TimerTask) Run() {
	tt.cron.Start()
	go func() {
		for {
			chosen, _, _ := reflect.Select(tt.cases)
			tw := tt.tws[chosen]
			tt.cases[chosen] = reflect.SelectCase{Dir: reflect.SelectRecv, Chan: reflect.ValueOf(time.NewTimer(tt.durations[tw]).C)}
			go func(tw *TaskContext) {
				if tt.runHandler != nil {
					tt.runHandler(tw.Name, time.Now().Add(tt.durations[tw]))
				}
				if tt.runContextHandler != nil {
					tt.runContextHandler(*tw)
				}
				err := tw.fuc(tw.args...)
				if tt.endHandler != nil {
					tt.endHandler(tw.Name, time.Now().Add(tt.durations[tw]))
				}
				if tt.endContextHandler != nil {
					tt.endContextHandler(*tw)
				}
				if err != nil {
					if tt.errHandler != nil {
						tt.errHandler(tw.Name, err)
					}
					if tt.errContextHandler != nil {
						tt.errContextHandler(*tw, err)
					}
				}
			}(tw)
		}
	}()
}
