package supervisor

import (
	"bufio"
	"fmt"
	"io"
	"os"
	"os/exec"
	"regexp"
	"strconv"
	"strings"
	"sync"
	"syscall"
	"time"

	"gitlab-ce.k8s.tools.vchangyi.com/cdp/cy-sdk-go/schedule"
	"go.uber.org/multierr"

	"gitlab-ce.k8s.tools.vchangyi.com/cdp/cy-sdk-go/exception"
)

const (
	READY    = "READY\n"
	RESULTOK = "RESULT 2\nOK"
)

type deamon struct {
	pid       string
	pname     string
	nextAlert int64
	alertTime int
}

func (d *deamon) resetState() {
	d.nextAlert = 0
	d.alertTime = 0
}
func (d *deamon) alert() {
	now := time.Now().Unix()
	if d.nextAlert < now {
		log(fmt.Sprintf("进程：pid:%s pname:%s 未存活", d.pid, d.pname))
		exception.SendAlert("supervisor_listener", fmt.Errorf("进程：pid:%s pname:%s 未存活", d.pid, d.pname))
		d.alertTime += 1
		if d.alertTime < 20 {
			d.nextAlert = now + 60
		} else if d.alertTime < 50 {
			d.nextAlert = now + 300
		} else if d.alertTime < 100 {
			d.nextAlert = now + 1800
		} else {
			d.nextAlert = now + 7200
		}
	}
	log(fmt.Sprintf("%s %s 下次报警时间：%d", d.pid, d.pname, d.nextAlert))
}

func NewListener() *Listener {
	return &Listener{
		deamons: make(map[string]*deamon, 0),
	}
}

//Listener supervisor 子进程状态监听
type Listener struct {
	deamons map[string]*deamon
	lock    sync.Mutex
}

func log(s string) {
	fmt.Fprintf(os.Stderr, fmt.Sprintf("info:%s\n", s))
	os.Stderr.Sync()
}

func (sl *Listener) state(s string) {

	fmt.Fprintf(os.Stdout, s)
	os.Stdout.Sync()
}

func (sl *Listener) registerDeamon(pname string, pid string) {
	sl.lock.Lock()
	defer sl.lock.Unlock()
	sl.deamons[pname] = &deamon{pid: pid, pname: pname, nextAlert: 0, alertTime: 0}
}

func (sl *Listener) unRegisterDeamon(pname string, pid string) {
	if _, ok := sl.deamons[pname]; ok {
		sl.lock.Lock()
		defer sl.lock.Unlock()
		delete(sl.deamons, pname)
	}
}
func (sl *Listener) init() {
	cmd := exec.Command("supervisorctl", "status", "")
	out, err := cmd.Output()
	if err != nil {
		log(err.Error())
	}
	outlines := strings.Split(string(out), "\n")
	sl.lock.Lock()
	defer sl.lock.Unlock()
	for _, outline := range outlines {
		spitem := strings.Split(outline, ",")
		pinfo := strings.Split(deleteExtraSpace(spitem[0]), " ")
		if len(pinfo) == 4 {
			if pinfo[1] == "RUNNING" {
				sl.deamons[pinfo[0]] = &deamon{pid: pinfo[3], pname: pinfo[0], nextAlert: 0, alertTime: 0}
			}
		}
	}
}
func (sl *Listener) checkDeamonState() {
	tt := schedule.NewTimerTask(schedule.WithErrHandler(func(name string, err error) {
		exception.SendAlert("supervisor_listener", fmt.Errorf("%s:%s", name, err.Error()))
	}), schedule.WithRunHandler(func(name string, nextAt time.Time) {
		// log(name)
	}))
	tt.Do(time.Now(), time.Second*10, func(args ...interface{}) error {
		sl.lock.Lock()
		defer sl.lock.Unlock()
		var err error
		for _, deamon := range sl.deamons {
			pid, e := strconv.Atoi(deamon.pid)
			if e != nil {
				err = multierr.Append(err, e)
				continue
			}
			p, err := os.FindProcess(pid)
			if err != nil {
				deamon.alert()
			} else {
				if p == nil {
					deamon.alert()
				} else {
					if err := p.Signal(syscall.Signal(0)); err != nil {
						deamon.alert()
					} else {
						deamon.resetState()
					}
				}
			}
		}
		return err
	}, "进程状态监测")
	tt.Run()
}

func (sl *Listener) recover(process, from, to string) {
	log(fmt.Sprintf("进程 %s 被拉起", process))
	exception.SendAlert("supervisor_listener", fmt.Errorf("进程 %s 拉起成功", process))
}

func (sl *Listener) panic(process, from, to string) {
	log(fmt.Sprintf("进程 %s 异常退出,请注意排查", process))
	exception.SendAlert("supervisor_listener", fmt.Errorf("进程 %s 异常退出,请注意排查", process))
}

func stringSliceContain(slice []string, value string) bool {
	for _, item := range slice {
		if item == value {
			return true
		}
	}
	return false
}

func deleteExtraSpace(s string) string {
	s1 := strings.Replace(s, "	", " ", -1)
	regstr := "\\s{2,}"
	reg, _ := regexp.Compile(regstr)
	s2 := make([]byte, len(s1))
	copy(s2, s1)
	spc_index := reg.FindStringIndex(string(s2))
	for len(spc_index) > 0 {
		s2 = append(s2[:spc_index[0]+1], s2[spc_index[1]:]...)
		spc_index = reg.FindStringIndex(string(s2))
	}
	return string(s2)
}

//Start 开始监听子进程状态
func (sl *Listener) Start() {
	sl.init()
	sl.checkDeamonState()
	lastSerial := "0"
	lastPoolSerial := "0"
	psExit := make(map[string]string, 0)
	var body [4096]byte
	reader := bufio.NewReader(os.Stdin)
	for {
		sl.state(READY)
		line, _, err := reader.ReadLine()
		if err != nil {
			exception.SendAlert("supervisor_listener", err)
			continue
		}
		log(string(line))
		header := make(map[string]string, 0)
		items := strings.Split(string(line), " ")
		for _, item := range items {
			kv := strings.Split(item, ":")
			header[kv[0]] = kv[1]
		}
		serial := header["serial"]
		poolSerial := header["poolserial"]
		bodylen, err := strconv.ParseFloat(header["len"], 32)
		if err != nil {
			exception.SendAlert("supervisor_listener", err)
			continue
		}
		_, err = io.ReadFull(reader, body[:int(bodylen)])
		if err != nil {
			exception.SendAlert("supervisor_listener", err)
			continue
		}
		data := body[:int(bodylen)]
		log("body+" + string(data))
		if serial == lastSerial && poolSerial == lastPoolSerial {
			sl.state(RESULTOK)
		}
		lastSerial = serial
		lastPoolSerial = poolSerial
		bodys := make(map[string]string, 0)
		items = strings.Split(string(data), " ")
		for _, item := range items {
			kv := strings.Split(item, ":")
			if len(kv) > 1 {
				bodys[kv[0]] = kv[1]
			}
		}
		process := bodys["processname"]
		fromState := bodys["from_state"]
		pid := bodys["pid"]
		toState := header["eventname"]
		log(fmt.Sprintf("\nprocessname:[%s] from_state:[%s] to_state:[%s] last_serial:[%s] last_pool_serial:[%s]\n", process, fromState, toState, lastSerial, lastPoolSerial))
		if stringSliceContain([]string{"PROCESS_STATE_RUNNING"}, toState) {
			sl.registerDeamon(process, pid)
			if _, ok := psExit[process]; ok {
				delete(psExit, process)
				sl.recover(process, fromState, toState)
			}
		}
		if stringSliceContain([]string{"PROCESS_STATE_STOPPED"}, toState) {
			sl.unRegisterDeamon(process, pid)
		}
		if stringSliceContain([]string{"PROCESS_STATE_EXITED", "PROCESS_STATE_FATAL", "PROCESS_STATE_BACKOFF"}, toState) {
			psExit[process] = ""
			sl.panic(process, fromState, toState)
		}
		sl.state(RESULTOK)
	}
}
