package uuid

import (
	"fmt"
	"net"
	"os"
	"time"

	"github.com/OneOfOne/xxhash"
	wuid "github.com/edwingeng/wuid/callback/wuid"
	rwuid "github.com/edwingeng/wuid/redis/wuid"
	"github.com/go-redis/redis"
	"github.com/spf13/viper"
)

// NewWuidWithHost 唯一标识生成器，主机标识
func NewWuidWithHost(tag string) (*wuid.WUID, error) {
	g := wuid.NewWUID(tag, nil)
	err := g.LoadH28WithCallback(func() (int64, func(), error) {
		addrs, err := net.InterfaceAddrs()
		if err != nil {
			return 0, nil, err
		}

		hostname, _ := os.Hostname()

		for _, address := range addrs {
			// 检查ip地址判断是否回环地址
			if ipnet, ok := address.(*net.IPNet); ok && !ipnet.IP.IsLoopback() {
				ipv4 := ipnet.IP.To4()
				if ipv4 != nil {
					hostname = hostname + ipv4.String()
					break
				}
			}
		}
		h := xxhash.New32()
		h.WriteString(hostname)
		return int64(h.Sum32() & 0x07FFFFFF), nil, nil
	})
	return g, err
}

// NewWuidWithRedis 唯一标识生成器，基于redis，key:生产id在redis中key
func NewWuidWithRedis(tag string, addr, pwd, key string, db int) (*rwuid.WUID, error) {
	newClient := func() (redis.Cmdable, bool, error) {
		client := redis.NewClient(&redis.Options{
			Addr:         addr,
			Password:     pwd,
			DB:           db,
			DialTimeout:  3 * time.Second,
			ReadTimeout:  3 * time.Second,
			WriteTimeout: 3 * time.Second,
			PoolSize:     10,
			PoolTimeout:  30 * time.Second,
		})
		return client, true, nil
	}
	g := rwuid.NewWUID(tag, nil)
	_ = g.LoadH28FromRedis(newClient, key)
	return g, nil
}

// NewWuidWithRedisViper 唯一标识生成器，基于redis，key:生产id在redis中key
func NewWuidWithRedisViper(vk string) (*rwuid.WUID, error) {
	v := viper.Sub(vk)
	if v == nil {
		return nil, fmt.Errorf("invalid config key: %s", vk)
	}

	newClient := func() (redis.Cmdable, bool, error) {
		client := redis.NewClient(&redis.Options{
			Addr:         v.GetString("redis_addr"),
			Password:     v.GetString("redis_pass"),
			DB:           v.GetInt("redis_index"),
			DialTimeout:  3 * time.Second,
			ReadTimeout:  3 * time.Second,
			WriteTimeout: 3 * time.Second,
			PoolSize:     10,
			PoolTimeout:  30 * time.Second,
		})
		return client, true, nil
	}
	v.SetDefault("tag", "default")
	g := rwuid.NewWUID(v.GetString("tag"), nil)
	v.SetDefault("wuid_key", "wuid")
	err := g.LoadH28FromRedis(newClient, v.GetString("wuid_key"))
	return g, err
}
