package client

import (
	"bytes"
	"encoding/json"
	"fmt"
	"io"
	"io/ioutil"
	"net/http"
	"net/url"
	"time"

	"github.com/glutwins/webclient"
	"gitlab-ce.k8s.tools.vchangyi.com/cdp/cy-sdk-go/wechat/cache"
	"gitlab-ce.k8s.tools.vchangyi.com/cdp/cy-sdk-go/wechat/crypt"
)

type WachatReq map[string]interface{}

type Client struct {
	*crypt.WechatConfig
	data     cache.Cache
	tkexpire bool
}

func NewClient(cfg *crypt.WechatConfig, d cache.Cache) *Client {
	c := &Client{WechatConfig: cfg, data: d}
	return c
}

type JsConfig struct {
	AppID     string
	TimeStamp int64
	NonceStr  string
	Signature string
}

type cacheClient struct {
	Token  accessToken
	Ticket resTicket
}

type tokenReq struct {
	Sign         string `json:"Auth-Sign"`
	OverdueToken string `json:"overdue_token"`
	Eid          int    `json:"enterprise_id"`
}

type tokenRsp struct {
	Token string `json:"access_token"`
}

//GetAccessToken 获取access_token
func (c *Client) GetAccessToken() (string, error) {
	var token accessToken
	if !c.tkexpire {
		if err := c.data.Get("token", &token); err == nil {
			if token.valid() {
				return token.AccessToken, nil
			}
		}
	}

	c.data.Lock("token")
	defer c.data.Unlock("token")

	url := fmt.Sprintf(accessTokenURL, c.AppID, c.AppSecret)
	resp, err := http.Get(url)
	if err != nil {
		return "", err
	}

	defer resp.Body.Close()
	if resp.StatusCode != http.StatusOK {
		return "", fmt.Errorf("GetAccessToken httpcode=%d", resp.StatusCode)
	}

	if b, err := ioutil.ReadAll(resp.Body); err != nil {
		return "", err
	} else if err = json.Unmarshal(b, &token); err != nil {
		return "", err
	}
	token.LastUpTs = time.Now().Unix()
	c.data.Set("token", token)
	c.tkexpire = false

	return token.AccessToken, token.Error()
}

// ExpireToken 使token失效
func (c *Client) ExpireToken() {
	c.tkexpire = true
}

func (c *Client) formatUrlWithAccessToken(base string, args ...interface{}) (string, error) {
	var token string
	var err error

	token, err = c.GetAccessToken()

	if err != nil {
		return "", err
	}
	return fmt.Sprintf(base, append([]interface{}{token}, args...)...), nil
}

func (c *Client) postJsonUrlFormat(req interface{}, res interface{}, url string, args ...interface{}) error {
	if uri, err := c.formatUrlWithAccessToken(url, args...); err != nil {
		return err
	} else if response, err := c.jsonPost(uri, req); err != nil {
		return err
	} else if err := json.Unmarshal(response, res); err != nil {
		return err
	}
	return nil
}

func (c *Client) jsonPost(uri string, obj interface{}) ([]byte, error) {
	var jsonData []byte
	var err error

	if str, ok := obj.(string); ok {
		jsonData = []byte(str)
	} else {
		jsonData, err = json.Marshal(obj)
		if err != nil {
			return nil, err
		}
	}
	return c.doPost(uri, "application/json;charset=utf-8", bytes.NewBuffer(jsonData))
}

func (c *Client) formPost(uri string, data map[string]string) ([]byte, error) {
	body := bytes.NewBuffer(nil)
	for k, v := range data {
		body.WriteString(k)
		body.WriteString("=")
		body.WriteString(url.QueryEscape(v))
		body.WriteString("&")
	}
	body.Truncate(body.Len() - 1)
	return c.doPost(uri, "application/x-www-form-urlencoded;charset=utf-8", body)
}

func (c *Client) doPost(uri, contentType string, r io.Reader) ([]byte, error) {
	response, err := http.Post(uri, contentType, r)
	if err != nil {
		return nil, err
	}
	defer response.Body.Close()

	if response.StatusCode != http.StatusOK {
		return nil, fmt.Errorf("http get error : uri=%v , statusCode=%v", uri, response.StatusCode)
	}
	return ioutil.ReadAll(response.Body)
}

func (c *Client) getJsonUrlFormat(res interface{}, url string, args ...interface{}) error {
	uri, err := c.formatUrlWithAccessToken(url, args...)
	if err != nil {
		return err
	}

	if b, err := webclient.DoGet(uri); err != nil {
		return err
	} else {
		return json.Unmarshal(b, res)
	}
}

//uri 为当前网页地址
func (js *Client) GetJsConfig(uri string) (config *JsConfig, err error) {
	config = new(JsConfig)
	var ticketStr string
	ticketStr, err = js.getTicket()
	if err != nil {
		return
	}

	nonceStr := crypt.RandomStr(16)
	timestamp := time.Now().Unix()
	str := fmt.Sprintf("jsapi_ticket=%s&noncestr=%s&timestamp=%d&url=%s", ticketStr, nonceStr, timestamp, uri)
	sigStr := crypt.Signature(str)

	config.AppID = js.AppID
	config.NonceStr = nonceStr
	config.TimeStamp = timestamp
	config.Signature = sigStr
	return
}

//getTicket 获取jsapi_tocket全局缓存
func (c *Client) getTicket() (string, error) {
	var ticket resTicket
	if err := c.data.Get("ticket", &ticket); err == nil {
		if ticket.valid() {
			return ticket.Ticket, nil
		}
	}

	c.data.Lock("ticket")
	defer c.data.Unlock("ticket")

	if err := c.getJsonUrlFormat(&ticket, getTicketURL); err != nil {
		return "", err
	}

	ticket.LastUpTs = time.Now().Unix()
	c.data.Set("ticket", ticket)

	return ticket.Ticket, ticket.Error()
}

func (c *Client) NewQrCode(sceneId interface{}, expire int) (*QrCode, error) {
	var msg string
	if _, ok := sceneId.(string); ok {
		if expire == 0 {
			msg = fmt.Sprintf(kQrStrLimitFormat, sceneId)
		} else {
			return nil, fmt.Errorf("not support expire str qrcode")
		}
	} else {
		if expire == 0 {
			msg = fmt.Sprintf(kQrIntLimitFormat, sceneId)
		} else {
			msg = fmt.Sprintf(kQrIntFormat, expire, sceneId)
		}
	}

	var qrcode = &QrCode{}
	if err := c.postJsonUrlFormat(msg, qrcode, qrcodeURL); err != nil {
		return nil, err
	}
	return qrcode, qrcode.Error()
}

type CustomerText struct {
	Content string `json:"content"`
}

type CustomerImage struct {
	MediaID string `json:"media_id"`
}

type CustomerMessage struct {
	ToUser  string         `json:"touser"`
	MsgType string         `json:"msgtype"`
	Text    *CustomerText  `json:"text,omitempty"`
	Image   *CustomerImage `json:"image,omitempty"`
}

func (c *Client) SendMessage(msg *CustomerMessage) error {
	resp := &CommonResp{}
	if err := c.postJsonUrlFormat(msg, resp, customerMsgURL); err != nil {
		return err
	}
	if resp.ErrCode == 40001 {
		if err := c.postJsonUrlFormat(msg, resp, customerMsgURL); err != nil {
			return err
		}
		return resp.Error()
	}
	return resp.Error()
}
