package wechat

import (
	"bytes"
	"encoding/xml"
	"io/ioutil"

	"gitlab-ce.k8s.tools.vchangyi.com/cdp/cy-sdk-go/wechat/crypt"
)

type MchBaseRes struct {
	ReturnCode string `xml:"return_code"`
	ReturnMsg  string `xml:"return_msg"`
	ResultCode string `xml:"result_code"`
	ErrCode    string `xml:"err_code"`
	ErrCodeDes string `xml:"err_code_des"`
}

func (r MchBaseRes) Error() string {
	if r.ReturnCode != "SUCCESS" {
		return r.ReturnMsg
	}

	if r.ResultCode != "SUCCESS" {
		return r.ErrCode + " " + r.ErrCodeDes
	}

	return ""
}

type MchTransRes struct {
	MchBaseRes
	AppId      string `xml:"mch_appid"`
	MchId      string `xml:"mchid"`
	DeviceInfo string `xml:"device_info"`
	NonceStr   string `xml:"nonce_str"`
	TradeNo    string `xml:"partner_trade_no"`
	PayNo      string `xml:"payment_no"`
	PayTime    string `xml:"payment_time"`
}

type MchTransferInfoRes struct {
	MchBaseRes
	TradeNo   string `xml:"partner_trade_no"`
	AppId     string `xml:"appid"`
	MchId     string `xml:"mchid"`
	PayNo     string `xml:"detail_id"`
	Status    string `xml:"status"`
	Reason    string `xml:"reason"`
	OpenId    string `xml:"openid"`
	Name      string `xml:"transfer_name"`
	Amount    int    `xml:"payment_amount"`
	TransTime string `xml:"transfer_time"`
	PayTime   string `xml:"payment_time"`
	Desc      string `xml:"desc"`
}

type mchRepackRes struct {
	MchBaseRes
	Sign    string `xml:"sign"`
	TradeNo string `xml:"mch_billno"`
	MchId   string `xml:"mch_id"`
	AppId   string `xml:"wxappid"`
	OpenId  string `xml:"re_openid"`
	Amount  int    `xml:"total_amount"`
	PayNo   string `xml:"send_listid"`
}

func (c *WechatClientV3) mchPost(url string, req map[string]interface{}, res error) error {
	sign := crypt.Md5Sign(req, c.apiSecret)
	req["sign"] = sign

	resp, err := c.secuClient.Post(url, "application/xml", bytes.NewReader([]byte(crypt.EncodeXml(req))))
	if err != nil {
		return err
	}
	b, err := ioutil.ReadAll(resp.Body)
	_ = resp.Body.Close()
	if err != nil {
		return err
	}

	if err = xml.Unmarshal(b, res); err != nil {
		return err
	}

	return nil
}

func (c *WechatClientV3) Transfer(openid string, tradeno string, amount int, desc string, checkName string, name string) (*MchTransRes, error) {
	val := make(map[string]interface{})
	val["mch_appid"] = c.AppID
	val["mchid"] = c.mchid
	val["nonce_str"] = crypt.RandomStr(10)
	val["partner_trade_no"] = tradeno
	val["openid"] = openid
	val["check_name"] = checkName
	val["re_user_name"] = name
	val["amount"] = amount
	val["desc"] = desc
	val["spbill_create_ip"] = ""

	res := &MchTransRes{}
	if err := c.mchPost("https://api.mch.weixin.qq.com/mmpaymkttransfers/promotion/transfers", val, res); err != nil {
		return nil, err
	}

	return res, nil
}

func (c *WechatClientV3) TransferInfo(tradeNo string) (*MchTransferInfoRes, error) {
	val := make(map[string]interface{})
	val["appid"] = c.AppID
	val["mch_id"] = c.mchid
	val["nonce_str"] = crypt.RandomStr(10)
	val["partner_trade_no"] = tradeNo

	res := &MchTransferInfoRes{}
	if err := c.mchPost("https://api.mch.weixin.qq.com/mmpaymkttransfers/gettransferinfo", val, res); err != nil {
		return nil, err
	}

	return res, nil
}

func (c *WechatClientV3) Redpack(openid, tradeno, mchname, wishing, act, remark, scene_id string, amount int, num int) (string, error) {
	val := make(map[string]interface{})
	val["nonce_str"] = crypt.RandomStr(10)
	val["mch_billno"] = tradeno
	val["mch_id"] = c.mchid
	val["wxappid"] = c.AppID
	val["send_name"] = mchname
	val["re_openid"] = openid
	val["total_amount"] = amount
	val["total_num"] = num
	val["wishing"] = wishing
	val["client_ip"] = ""
	val["act_name"] = act
	val["remark"] = remark

	if amount > 20000 {
		val["scene_id"] = scene_id
	}

	var res mchRepackRes
	if err := c.mchPost("https://api.mch.weixin.qq.com/mmpaymkttransfers/sendredpack", val, &res); err != nil {
		return "", nil
	}

	return res.PayNo, nil
}

func (c *WechatClientV3) GroupRedpack(openid, tradeno, mchname, wishing, act, remark, scene_id string, amount int, num int) (string, error) {
	val := make(map[string]interface{})
	val["nonce_str"] = crypt.RandomStr(10)
	val["mch_billno"] = tradeno
	val["mch_id"] = c.mchid
	val["wxappid"] = c.AppID
	val["send_name"] = mchname
	val["re_openid"] = openid
	val["total_amount"] = amount
	val["total_num"] = num
	val["wishing"] = wishing
	val["client_ip"] = ""
	val["act_name"] = act
	val["remark"] = remark
	val["amt_type"] = "ALL_RAND"
	val["scene_id"] = scene_id

	var res mchRepackRes
	if err := c.mchPost("https://api.mch.weixin.qq.com/mmpaymkttransfers/sendgroupredpack", val, &res); err != nil {
		return "", nil
	}

	return res.PayNo, nil
}
