package wechat

import (
	"fmt"
	"strconv"
	"time"
)

var wechatTimeFormat = "2006-01-02T15:04:05-07:00"

// WechatTime 微信格式时间序列化
type WechatTime time.Time

// UnmarshalJSON 反序列化
func (t *WechatTime) UnmarshalJSON(data []byte) error {
	if len(data) <= 2 {
		return nil
	}
	tm, err := time.Parse(wechatTimeFormat, string(data[1:len(data)-1]))
	if err == nil {
		*t = WechatTime(tm)
	}
	return err
}

// MarshalJSON 序列化
func (t WechatTime) MarshalJSON() ([]byte, error) {
	return []byte(time.Time(t).Format(`"` + wechatTimeFormat + `"`)), nil
}

type SingleTypeMsg struct {
	SinglePriceMax int64 `json:"single_price_max"`
	CutToPrice     int64 `json:"cut_to_price"`
}

type FixedValueStockMsg struct {
	CouponAmount   int64 `json:"coupon_amount"`
	TransactionMin int64 `json:"transaction_minimum"`
}

type CouponConsumeInfo struct {
	ConsumeTime   WechatTime               `json:"consume_time"`
	ConsumeMchID  string                   `json:"consume_mchid"`
	TransactionID string                   `json:"transaction_id"`
	GoodsDetail   []map[string]interface{} `json:"goods_detail"`
}

type CouponEntity struct {
	StockCreatorMchID string              `json:"stock_creator_mchid"`
	StockID           string              `json:"stock_id"`
	CouponID          string              `json:"coupon_id"`
	CutToMessage      *SingleTypeMsg      `json:"cut_to_message,omitempty"`
	CouponName        string              `json:"coupon_name"`
	Status            string              `json:"status"`
	Description       string              `json:"description"`
	CreateTime        WechatTime          `json:"create_time"`
	CouponType        string              `json:"coupon_type"`
	NoCash            bool                `json:"no_cash"`
	AvlBeginTime      WechatTime          `json:"available_begin_time"`
	AvlEndTime        WechatTime          `json:"available_end_time"`
	SingleItem        bool                `json:"singleitem"`
	NormalCouponInfo  *FixedValueStockMsg `json:"normal_coupon_information,omitempty"`
	ConsumeInfo       *CouponConsumeInfo  `json:"consume_information,omitempty"`
}

type StockUseRule struct {
	MaxCoupons        int                 `json:"max_coupons"`
	MaxAmount         int                 `json:"max_amount"`
	MaxAmountByDay    int                 `json:"max_amount_by_day"`
	FixedNormalCoupon *FixedValueStockMsg `json:"fixed_normal_coupon,omitempty"`
	MaxCouponsPerUser int                 `json:"max_coupons_per_user"`
	CouponType        string              `json:"coupon_type"`
	CombineUse        bool                `json:"combine_use"`
}

type StockEntity struct {
	StockCreatorMchID string         `json:"stock_creator_mchid"`
	StockID           string         `json:"stock_id"`
	StockName         string         `json:"stock_name"`
	Status            string         `json:"status"`
	CreateTime        string         `json:"create_time"`
	Description       string         `json:"description"`
	StockUseRule      *StockUseRule  `json:"stock_use_rule,omitempty"`
	AvlBeginTime      WechatTime     `json:"available_begin_time"`
	AvlEndTime        WechatTime     `json:"available_end_time"`
	NoCash            bool           `json:"no_cash"`
	StartTime         WechatTime     `json:"start_time"`
	StopTime          WechatTime     `json:"stop_time"`
	CutToMessage      *SingleTypeMsg `json:"cut_to_message,omitempty"`
	SingleItem        bool           `json:"singleitem"`
	StockType         string         `json:"stock_type"`
}

func (stock StockEntity) NewCoupon(couponID string) *CouponEntity {
	entity := &CouponEntity{}
	entity.AvlBeginTime = stock.AvlBeginTime
	entity.AvlEndTime = stock.AvlEndTime
	entity.CouponID = couponID
	entity.CouponName = stock.StockName
	entity.CouponType = stock.StockType
	entity.CreateTime = WechatTime(time.Now())
	entity.CutToMessage = stock.CutToMessage
	entity.Description = stock.Description
	entity.NoCash = stock.NoCash
	entity.NormalCouponInfo = stock.StockUseRule.FixedNormalCoupon
	entity.SingleItem = stock.SingleItem
	entity.Status = "SENDED"
	entity.StockCreatorMchID = stock.StockCreatorMchID
	entity.StockID = stock.StockID
	return entity
}

type QueryByUseMchResponse struct {
	TotalCount int             `json:"total_count"`
	Data       []*CouponEntity `json:"data"`
}

func (resp QueryByUseMchResponse) HasStock(stockid string) bool {
	for _, stock := range resp.Data {
		if stock.StockID == stockid {
			return true
		}
	}
	return false
}

type StockListResponse struct {
	TotalCount int            `json:"total_count"`
	Data       []*StockEntity `json:"data"`
}

type StockUseFlowResponse struct {
	Url       string `json:"url"`
	HashValue string `json:"hash_value"`
	HashType  string `json:"hash_type"`
}

type ErrResponse struct {
	Link    string `json:"-"`
	AppID   string `json:"-"`
	Code    string `json:"code"`
	Message string `json:"message"`
}

func (resp ErrResponse) Error() string {
	return fmt.Sprintf("%s request %s (%s: %s)", resp.AppID, resp.Link, resp.Code, resp.Message)
}

// WechatOption 微信appOnShow参数结构体
type WechatOption struct {
	Path         string      `json:"path"`
	Scene        int         `json:"scene"`
	Query        interface{} `json:"query"`
	ShareTicket  string      `json:"shareTicket"`
	ReferrerInfo interface{} `json:"referrerInfo"`
}

func (opt *WechatOption) GetExtraData() (string, map[string]interface{}) {
	if opt.ReferrerInfo == nil {
		return "", nil
	}

	refer, ok := opt.ReferrerInfo.(map[string]interface{})
	if !ok {
		return "", nil
	}

	var appid string
	var query map[string]interface{}
	if v, ok := refer["appId"].(string); ok {
		appid = v
	}
	if v, ok := refer["extraData"].(map[string]interface{}); ok {
		query = v
	}
	return appid, query
}

func (opt *WechatOption) GetQueryString(key string) string {
	query, ok := opt.Query.(map[string]interface{})
	if !ok {
		return ""
	}
	v, ok := query[key].(string)
	if !ok {
		return ""
	}
	return v
}

func (opt *WechatOption) GetQueryInt(key string) int {
	query, ok := opt.Query.(map[string]interface{})
	if !ok {
		return 0
	}
	if v, ok := query[key].(float64); ok {
		return int(v)
	}

	if v, ok := query[key].(int); ok {
		return v
	}

	if v, ok := query[key].(string); ok {
		iv, _ := strconv.Atoi(v)
		return iv
	}
	return 0
}

// StockWithValue 指定面额券
type StockWithValue struct {
	StockID       string // 批次号
	CrtMchID      string // 创建商户号
	AppID         string // appid
	CouponValue   int    // 金额
	CouponMinimum int    // 门槛
}
type MchStockResp struct {
	Data       []*MchStock `json:"data"`
	TotalCount int         `json:"total_count"`
	Limit      int         `json:"limit"`
	Offset     int         `json:"offset"`
}
type MchStock struct {
	BelongMerchant     string `json:"belong_merchant"`
	StockName          string `json:"stock_name"`
	Comment            string `json:"comment"`
	GoodsName          string `json:"goods_name"`
	StockType          string `json:"stock_type"`
	Transferable       string `json:"transferable"`
	Shareable          string `json:"shareable"`
	CouponState        string `json:"coupon_state"`
	DisplayPatternInfo struct {
		MerchantLogoURL string `json:"merchant_logo_url"`
		MerchantName    string `json:"merchant_name"`
		BackgroundColor string `json:"background_color"`
		CouponImageURL  string `json:"coupon_image_url"`
	} `json:"display_pattern_info"`
	CouponUseRule struct {
		CouponAvailableTime struct {
			AvailableBeginTime       time.Time `json:"available_begin_time"`
			AvailableEndTime         time.Time `json:"available_end_time"`
			AvailableDayAfterReceive int       `json:"available_day_after_receive"`
			AvailableWeek            struct {
				WeekDay          []string `json:"week_day"`
				AvailableDayTime []struct {
					BeginTime int `json:"begin_time"`
					EndTime   int `json:"end_time"`
				} `json:"available_day_time"`
			} `json:"available_week"`
			IrregularyAvaliableTime []struct {
				BeginTime time.Time `json:"begin_time"`
				EndTime   time.Time `json:"end_time"`
			} `json:"irregulary_avaliable_time"`
		} `json:"coupon_available_time"`
		FixedNormalCoupon struct {
			DiscountAmount     int `json:"discount_amount"`
			TransactionMinimum int `json:"transaction_minimum"`
		} `json:"fixed_normal_coupon"`
		DiscountCoupon struct {
			DiscountPercent    int `json:"discount_percent"`
			TransactionMinimum int `json:"transaction_minimum"`
		} `json:"discount_coupon"`
		ExchangeCoupon struct {
			ExchangePrice      int `json:"exchange_price"`
			TransactionMinimum int `json:"transaction_minimum"`
		} `json:"exchange_coupon"`
		UseMethod         string `json:"use_method"`
		MiniProgramsAppid string `json:"mini_programs_appid"`
		MiniProgramsPath  string `json:"mini_programs_path"`
	} `json:"coupon_use_rule"`
	CustomEntrance struct {
		MiniProgramsInfo struct {
			MiniProgramsAppid string `json:"mini_programs_appid"`
			MiniProgramsPath  string `json:"mini_programs_path"`
			EntranceWords     string `json:"entrance_words"`
			GuidingWords      string `json:"guiding_words"`
		} `json:"mini_programs_info"`
		Appid   string `json:"appid"`
		HallID  string `json:"hall_id"`
		StoreID string `json:"store_id"`
	} `json:"custom_entrance"`
	CouponCode         string    `json:"coupon_code"`
	StockID            string    `json:"stock_id"`
	AvailableStartTime time.Time `json:"available_start_time"`
	ExpireTime         time.Time `json:"expire_time"`
	ReceiveTime        time.Time `json:"receive_time"`
	SendRequestNo      string    `json:"send_request_no"`
	UseRequestNo       string    `json:"use_request_no"`
	UseTime            time.Time `json:"use_time"`
}
