package wechat

import (
	"bytes"
	"crypto"
	"crypto/rand"
	"crypto/rsa"
	"crypto/sha256"
	"crypto/tls"
	"crypto/x509"
	"encoding/base64"
	"encoding/json"
	"encoding/pem"
	"errors"
	"fmt"
	"io/ioutil"
	"os"

	"net/http"
	"net/url"
	"time"

	jsoniter "github.com/json-iterator/go"
	"github.com/juju/ratelimit"
)

var host = "https://api.mch.weixin.qq.com"

// SetHost 更新服务器
func SetHost(h string) {
	host = h
}

// RateConfig 频率控制
type RateConfig struct {
	Cap          int
	WaitDuration int
}

// ClientV3Config 配置
type ClientV3Config struct {
	MchID        string `db:"mch_id"`
	ServiceMchID string `db:"service_mch_id"`
	PemKey       []byte `db:"pemkey"`
	PemCert      []byte `db:"pemcert"`
	SerialNO     string `db:"serial_no"`
	AppID        string `db:"appid"`
	APISecret    string `db:"api_secret"`
}

// WechatClientV3 微信V3接口
type WechatClientV3 struct {
	pemkey       *rsa.PrivateKey
	mchid        string
	serviceMchID string
	AppID        string
	serialno     string
	apiSecret    string

	authClient *http.Client
	secuClient *http.Client
	rate       *ratelimit.Bucket
	wait       func() error
}

// NewWechatClientV3WithPem 创建v3接口
func NewWechatClientV3WithPem(conf *ClientV3Config, rate *RateConfig) (*WechatClientV3, error) {
	block, _ := pem.Decode(conf.PemKey)
	if block == nil || block.Type != "PRIVATE KEY" {
		return nil, errors.New("invalid private key")
	}

	pri, err := x509.ParsePKCS8PrivateKey(block.Bytes)
	if err != nil {
		return nil, err
	}

	c := &WechatClientV3{
		mchid:        conf.MchID,
		serviceMchID: conf.ServiceMchID,
		AppID:        conf.AppID,
		pemkey:       pri.(*rsa.PrivateKey),
		serialno:     conf.SerialNO,
		apiSecret:    conf.APISecret,

		authClient: &http.Client{Transport: &http.Transport{MaxIdleConnsPerHost: 100}},
	}

	cert, err := tls.X509KeyPair(conf.PemCert, conf.PemKey)
	if err != nil {
		return nil, err
	}
	clientCertPool := x509.NewCertPool()
	if !clientCertPool.AppendCertsFromPEM(conf.PemCert) {
		return nil, fmt.Errorf("failed to parse root certificate")
	}

	tlsconf := &tls.Config{
		RootCAs:            clientCertPool,
		Certificates:       []tls.Certificate{cert},
		InsecureSkipVerify: true,
	}
	c.secuClient = &http.Client{Transport: &http.Transport{TLSClientConfig: tlsconf}}

	if rate != nil && rate.Cap > 0 {
		c.rate = ratelimit.NewBucket(time.Second/time.Duration(rate.Cap), int64(rate.Cap))
		if rate.WaitDuration != 0 {
			c.wait = func() error {
				if !c.rate.WaitMaxDuration(1, time.Second*5) {
					return errors.New("rate limit")
				}
				return nil
			}
		} else {
			c.wait = func() error {
				c.rate.Wait(1)
				return nil
			}
		}
	} else {
		c.wait = func() error {
			return nil
		}
	}
	return c, nil
}

// NewWechatClientV3 创建V3接口
func NewWechatClientV3(mchid string, serviceMchID, appid string, pemfile string, serialno string, rate *RateConfig) (*WechatClientV3, error) {
	if serialno == "" {
		return nil, errors.New("serialno is empty")
	}
	data, err := ioutil.ReadFile(pemfile)
	if err != nil {
		return nil, err
	}

	return NewWechatClientV3WithPem(&ClientV3Config{MchID: mchid, ServiceMchID: serviceMchID, PemKey: data, AppID: appid, SerialNO: serialno}, rate)
}

func (c *WechatClientV3) doRequest(link string, req *http.Request) ([]byte, error) {
	resp, err := c.authClient.Do(req)
	if err != nil {
		return nil, &ErrResponse{Link: req.URL.String(), AppID: c.AppID, Code: "REQ_ERR", Message: err.Error()}
	}

	b, err := ioutil.ReadAll(resp.Body)
	resp.Body.Close()
	if err != nil {
		return nil, &ErrResponse{Link: req.URL.Path, AppID: c.AppID, Code: "READ_ERR", Message: err.Error()}
	}

	if resp.StatusCode != http.StatusOK {
		if len(b) > 0 {
			eret := &ErrResponse{Link: req.URL.Path, AppID: c.AppID}
			if err = jsoniter.ConfigCompatibleWithStandardLibrary.Unmarshal(b, eret); err != nil {
				eret.Code = resp.Status
				eret.Message = err.Error() + "(" + string(b) + ")"
			}
			return nil, eret
		} else {
			return nil, &ErrResponse{Link: req.URL.Path, AppID: c.AppID, Code: resp.Status}
		}
	}
	return b, nil
}

// DoGet Get请求
func (c *WechatClientV3) DoGet(urlstr string) ([]byte, error) {
	if err := c.wait(); err != nil {
		return nil, err
	}
	token, err := c.authrization("GET", urlstr, nil)
	if err != nil {
		return nil, err
	}

	req, err := http.NewRequest("GET", urlstr, nil)
	if err != nil {
		return nil, err
	}

	req.Header.Add("Accept", "application/json")
	req.Header.Add("Content-Type", "application/json")
	req.Header.Add("Authorization", token)
	return c.doRequest(urlstr, req)
}

// DoPost Post请求
func (c *WechatClientV3) DoPost(urlstr string, body interface{}) ([]byte, error) {
	if err := c.wait(); err != nil {
		return nil, err
	}
	token, err := c.authrization("POST", urlstr, body)
	if err != nil {
		return nil, err
	}

	b, err := jsoniter.ConfigCompatibleWithStandardLibrary.Marshal(body)
	if err != nil {
		return nil, err
	}
	req, err := http.NewRequest("POST", urlstr, bytes.NewReader(b))
	if err != nil {
		return nil, err
	}

	req.Header.Add("Accept", "application/json")
	req.Header.Add("Content-Type", "application/json")
	req.Header.Add("Authorization", token)
	req.Header.Add("Idempotency-Key", RandString(16))

	return c.doRequest(urlstr, req)
}

func (c *WechatClientV3) authrization(method string, urlstr string, body interface{}) (string, error) {
	buf := bytes.NewBuffer(nil)
	buf.WriteString(method + "\n")
	uri, _ := url.Parse(urlstr)
	path := uri.Path
	if uri.RawQuery != "" {
		path = path + "?" + uri.RawQuery
	}
	buf.WriteString(path + "\n")
	ts := time.Now().Unix()
	buf.WriteString(fmt.Sprintf("%d\n", ts))
	nonce := RandString(16)
	buf.WriteString(nonce + "\n")
	if body != nil {
		b, _ := jsoniter.ConfigCompatibleWithStandardLibrary.Marshal(body)
		buf.Write(b)
	}
	buf.WriteString("\n")

	hash := sha256.New()
	hash.Write([]byte(buf.Bytes()))
	b, err := rsa.SignPKCS1v15(rand.Reader, c.pemkey, crypto.SHA256, hash.Sum(nil))
	if err != nil {
		return "", err
	}
	sign := base64.StdEncoding.EncodeToString(b)
	return fmt.Sprintf(`WECHATPAY2-SHA256-RSA2048 mchid="%s",nonce_str="%s",timestamp="%d",serial_no="%s",signature="%s"`, c.mchid, nonce, ts, c.serialno, sign), nil
}

// RandString 产生随机字符串
func RandString(n int) string {
	const alphanum = "0123456789" +
		"ABCDEFGHIJKLMNOPQRSTUVWXYZ" +
		"abcdefghijklmnopqrstuvwxyz"

	var bytes = make([]byte, n)
	rand.Read(bytes)
	for i, b := range bytes {
		bytes[i] = alphanum[b%byte(len(alphanum))]
	}
	return string(bytes)
}

// https://pay.weixin.qq.com/wiki/doc/apiv3/convention.php?chapter=13_5&index=5#schemaCouponEntity

func (c *WechatClientV3) getCrtMchID(crtMchID string) string {
	if len(crtMchID) > 0 {
		return crtMchID
	}
	if len(c.serviceMchID) > 0 {
		return c.serviceMchID
	}
	return c.mchid
}

// GetCopuonByUserWithAvlMch 获取用户的代金券(只能使用核销商户号查询，否则返回空)
func (c *WechatClientV3) GetCopuonByUserWithAvlMch(openid string) (*QueryByUseMchResponse, error) {
	link := fmt.Sprintf("%s/v3/marketing/favor/users/%s/coupons?appid=%s&available_mchid=%s", host, openid, c.AppID, c.mchid)
	b, err := c.DoGet(link)
	if err != nil {
		return nil, err
	}
	resp := &QueryByUseMchResponse{}
	err = jsoniter.ConfigCompatibleWithStandardLibrary.Unmarshal(b, resp)
	return resp, err
}

// GetCopuonByUserWithCrtMch 获取用户的代金券(制券商户号)
func (c *WechatClientV3) GetCopuonByUserWithCrtMch(openid string, offset, limit int) (*QueryByUseMchResponse, error) {
	link := fmt.Sprintf("%s/v3/marketing/favor/users/%s/coupons?appid=%s&creator_mchid=%s&status=SENDED&offset=%d&limit=%d", host, openid, c.AppID, c.mchid, offset, limit)
	b, err := c.DoGet(link)

	if err != nil {
		return nil, err
	}
	resp := &QueryByUseMchResponse{}
	err = jsoniter.ConfigCompatibleWithStandardLibrary.Unmarshal(b, resp)

	return resp, err
}

// GetCouponByStockWithCrtMch 查询用户在制券商户号下的某个批次的全部券
func (c *WechatClientV3) GetCouponByStockWithCrtMch(openid string, stockid string) ([]*CouponEntity, error) {
	link := fmt.Sprintf("%s/v3/marketing/favor/users/%s/coupons?appid=%s&creator_mchid=%s&status=SENDED&offset=0&limit=100&stock_id=%s", host, openid, c.AppID, c.mchid, stockid)
	b, err := c.DoGet(link)

	if err != nil {
		return nil, err
	}
	resp := &QueryByUseMchResponse{}
	err = jsoniter.ConfigCompatibleWithStandardLibrary.Unmarshal(b, resp)

	return resp.Data, err
}

// GetCopuonByUserWithCrtMchAll 获取用户的所有代金券(制券商户号)
func (c *WechatClientV3) GetCopuonByUserWithCrtMchAll(openid string) ([]*CouponEntity, error) {
	offset, limit := 0, 100
	var coupons []*CouponEntity
	for {
		resp, err := c.GetCopuonByUserWithCrtMch(openid, offset, limit)
		if err != nil {
			return coupons, err
		}
		coupons = append(coupons, resp.Data...)
		if resp.TotalCount == len(coupons) || len(resp.Data) < limit {
			return coupons, nil
		}
	}
}

// IterCopuonByUserWithCrtMch 遍历获取用户的所有代金券(制券商户号)
func (c *WechatClientV3) IterCopuonByUserWithCrtMch(openid string, fn func([]*CouponEntity) bool) (bool, error) {
	offset, limit, count := 0, 100, 0
	for {
		resp, err := c.GetCopuonByUserWithCrtMch(openid, offset, limit)
		if err != nil {
			return false, err
		}

		if fn(resp.Data) {
			return true, nil
		}
		count += len(resp.Data)
		if resp.TotalCount == count || len(resp.Data) < limit {
			return false, nil
		}
	}
}

// GetCopuonByUserWithSenderMch 获取用户的代金券(制券商户号)
func (c *WechatClientV3) GetCopuonByUserWithSenderMch(openid string, offset, limit int) (*QueryByUseMchResponse, error) {
	link := fmt.Sprintf("%s/v3/marketing/favor/users/%s/coupons?appid=%s&sender_mchid=%s&status=SENDED&offset=%d&limit=%d", host, openid, c.AppID, c.mchid, offset, limit)
	b, err := c.DoGet(link)

	if err != nil {
		return nil, err
	}
	resp := &QueryByUseMchResponse{}
	err = jsoniter.ConfigCompatibleWithStandardLibrary.Unmarshal(b, resp)

	return resp, err
}

// GetCopuonByUserWithSenderMchAll 获取用户的所有代金券(制券商户号)
func (c *WechatClientV3) GetCopuonByUserWithSenderMchAll(openid string) ([]*CouponEntity, error) {
	offset, limit := 0, 100
	var coupons []*CouponEntity
	for {
		resp, err := c.GetCopuonByUserWithSenderMch(openid, offset, limit)
		if err != nil {
			return coupons, err
		}
		coupons = append(coupons, resp.Data...)
		if resp.TotalCount == len(coupons) || len(resp.Data) < limit {
			return coupons, nil
		}
	}
}

// IterCopuonByUserWithSenderMch 遍历获取用户的所有代金券(制券商户号)
func (c *WechatClientV3) IterCopuonByUserWithSenderMch(openid string, fn func([]*CouponEntity) bool) (bool, error) {
	offset, limit, count := 0, 100, 0
	for {
		resp, err := c.GetCopuonByUserWithSenderMch(openid, offset, limit)
		if err != nil {
			return false, err
		}

		if fn(resp.Data) {
			return true, nil
		}
		count += len(resp.Data)
		if resp.TotalCount == count || len(resp.Data) < limit {
			return false, nil
		}
	}
}

// GetCouponByID 根据优惠券标识查询详情
func (c *WechatClientV3) GetCouponByID(openid string, couponID string) (*CouponEntity, error) {
	link := fmt.Sprintf("%s/v3/marketing/favor/users/%s/coupons/%s?appid=%s", host, openid, couponID, c.AppID)
	b, err := c.DoGet(link)

	if err != nil {
		return nil, err
	}
	resp := &CouponEntity{}
	err = jsoniter.ConfigCompatibleWithStandardLibrary.Unmarshal(b, resp)

	return resp, err
}

// SendCoupon 发放代金券
func (c *WechatClientV3) SendCoupon(openid, stock, crtMchID, appid string) (string, error) {
	rawurl := fmt.Sprintf("%s/v3/marketing/favor/users/%s/coupons", host, openid)
	requestno := fmt.Sprintf("%s%s%s", c.mchid, time.Now().Format("20060102150405"), RandString(8))
	if appid == "" {
		appid = c.AppID
	}
	b, err := c.DoPost(rawurl, map[string]interface{}{
		"stock_id":            stock,
		"out_request_no":      requestno,
		"appid":               appid,
		"stock_creator_mchid": c.getCrtMchID(crtMchID),
	})
	if err != nil {
		return "", err
	}
	resp := struct {
		Code     string `json:"code"`
		Message  string `json:"message"`
		CouponID string `json:"coupon_id"`
	}{}
	if err := jsoniter.ConfigCompatibleWithStandardLibrary.Unmarshal(b, &resp); err != nil {
		return "", err
	}

	if resp.CouponID == "" {
		return "", errors.New(resp.Message)
	}
	return resp.CouponID, nil
}

// SendCoupon 发放代金券
func (c *WechatClientV3) SendCouponWithValue(openid string, stock *StockWithValue) (string, error) {
	rawurl := fmt.Sprintf("%s/v3/marketing/favor/users/%s/coupons", host, openid)
	requestno := fmt.Sprintf("%s%s%s", c.mchid, time.Now().Format("20060102150405"), RandString(8))
	if stock.AppID == "" {
		stock.AppID = c.AppID
	}
	if stock.CrtMchID == "" {
		stock.CrtMchID = c.mchid
	}
	b, err := c.DoPost(rawurl, map[string]interface{}{
		"stock_id":            stock.StockID,
		"out_request_no":      requestno,
		"appid":               stock.AppID,
		"stock_creator_mchid": stock.CrtMchID,
		"coupon_value":        stock.CouponValue,
		"coupon_minimum":      stock.CouponMinimum,
	})
	if err != nil {
		return "", err
	}
	resp := struct {
		Code     string `json:"code"`
		Message  string `json:"message"`
		CouponID string `json:"coupon_id"`
	}{}
	if err := jsoniter.ConfigCompatibleWithStandardLibrary.Unmarshal(b, &resp); err != nil {
		return "", err
	}

	if resp.CouponID == "" {
		return "", errors.New(resp.Message)
	}
	return resp.CouponID, nil
}
func (c *WechatClientV3) GetStockByID(stockid string, crtMchID string) (*StockEntity, error) {
	link := fmt.Sprintf("%s/v3/marketing/favor/stocks/%s?stock_creator_mchid=%s", host, stockid, c.getCrtMchID(crtMchID))
	b, err := c.DoGet(link)
	if err != nil {
		return nil, err
	}
	resp := &StockEntity{}
	err = jsoniter.ConfigCompatibleWithStandardLibrary.Unmarshal(b, resp)

	return resp, err
}

const (
	// StockTypeFavor 代金券
	StockTypeFavor = "FAVOR_STOCK"
	// StockTypeBusiFavor 商家券
	StockTypeBusiFavor = "BUSIFAVOR_STOCK"
)

// PartnerShip 建立合作关系
func (c *WechatClientV3) PartnerShip(stockType string, mchid, stockid string) (map[string]interface{}, error) {
	link := fmt.Sprintf("%s/v3/marketing/partnerships/build", host)
	b, err := c.DoPost(link, map[string]interface{}{
		"partner": map[string]interface{}{
			"merchant_id": mchid,
			"type":        "MERCHANT",
		},

		"authorized_data": map[string]interface{}{
			"business_type": stockType,
			"stock_id":      stockid,
		},
	})
	if err != nil {
		return nil, err
	}
	var resp map[string]interface{}
	if err := json.Unmarshal(b, &resp); err != nil {
		return nil, err
	}

	return resp, err
}

// DownloadStockUse 下载批次号核销明细及退款明细
func (c *WechatClientV3) DownloadStockUse(stockid string, useFile string, refundFile string) error {
	if _, err := os.Stat(useFile); err == nil {
		return nil
	}

	resp, err := c.GetStockRefundFlow(stockid)
	if err == nil {
		res, err := c.DoGet(resp.Url)
		if err != nil {
			return err
		}
		if err = ioutil.WriteFile(refundFile, res, 0644); err != nil {
			return err
		}
	} else {
		if werr, ok := err.(*ErrResponse); !ok || werr.Code != "APPID_MCHID_NOT_MATCH" {
			return err
		}
	}

	resp, err = c.GetStockUseFlow(stockid)
	if err == nil {
		res, err := c.DoGet(resp.Url)
		if err != nil {
			return err
		}
		return ioutil.WriteFile(useFile, res, 0644)
	}
	if werr, ok := err.(*ErrResponse); ok && werr.Code == "APPID_MCHID_NOT_MATCH" {
		return ioutil.WriteFile(useFile, []byte("`总条数\n0"), 0644)
	}
	return err
}

func (c *WechatClientV3) GetStockList(crtMchID string, offset int) ([]*StockEntity, error) {
	link := fmt.Sprintf("%s/v3/marketing/favor/stocks?offset=%d&limit=10&stock_creator_mchid=%s",
		host, offset, c.getCrtMchID(crtMchID))
	b, err := c.DoGet(link)
	if err != nil {
		return nil, err
	}
	resp := &StockListResponse{}
	err = jsoniter.ConfigCompatibleWithStandardLibrary.Unmarshal(b, resp)

	return resp.Data, err
}

// GetStockUseFlow 获取指定批次号的核销明细
func (c *WechatClientV3) GetStockUseFlow(stockid string) (*StockUseFlowResponse, error) {
	link := fmt.Sprintf("%s/v3/marketing/favor/stocks/%s/use-flow", host, stockid)
	b, err := c.DoGet(link)
	if err != nil {
		if werr, ok := err.(*ErrResponse); ok && werr.Code == "FREQUENCY_LIMITED" {
			time.Sleep(time.Second * 3)
			return c.GetStockUseFlow(stockid)
		}
		return nil, err
	}
	resp := &StockUseFlowResponse{}
	err = jsoniter.ConfigCompatibleWithStandardLibrary.Unmarshal(b, resp)
	return resp, err
}

// GetStockRefundFlow 获取指定批次号的退款明细
func (c *WechatClientV3) GetStockRefundFlow(stockid string) (*StockUseFlowResponse, error) {
	link := fmt.Sprintf("%s/v3/marketing/favor/stocks/%s/refund-flow", host, stockid)
	b, err := c.DoGet(link)
	if err != nil {
		if werr, ok := err.(*ErrResponse); ok && werr.Code == "FREQUENCY_LIMITED" {
			time.Sleep(time.Second * 3)
			return c.GetStockRefundFlow(stockid)
		}
		return nil, err
	}
	resp := &StockUseFlowResponse{}
	err = jsoniter.ConfigCompatibleWithStandardLibrary.Unmarshal(b, resp)
	return resp, err
}

// GetMchStockByUser 获取用户所有商家券
func (c *WechatClientV3) GetMchStockByUser(openid string, offset, limit int) (*MchStockResp, error) {
	link := fmt.Sprintf("%s/v3/marketing/busifavor/users/%s/coupons?appid=%s&offset=%d&limit=%d&creator_merchant=%s&coupon_state=SENDED", host, openid, c.AppID, offset, limit, c.mchid)
	b, err := c.DoGet(link)
	if err != nil {
		return nil, err
	}
	var resp *MchStockResp
	if err := json.Unmarshal(b, &resp); err != nil {
		return nil, err
	}
	return resp, err
}

// GetUserMchStockAll 遍历获取用户的所有的商家券
func (c *WechatClientV3) GetUserMchStockAll(openid string) ([]*MchStock, error) {
	offset, limit := 0, 30
	var coupons []*MchStock
	for {
		resp, err := c.GetMchStockByUser(openid, offset, limit)
		if err != nil {
			return nil, err
		}
		coupons = append(coupons, resp.Data...)
		if resp.TotalCount == len(coupons) || len(resp.Data) < limit {
			return coupons, nil
		}
	}
}

//CancalUserMchStock 核销商家券
func (c *WechatClientV3) CancalUserMchStock(couponCode string) error {
	rawurl := fmt.Sprintf("%s/v3/marketing/busifavor/coupons/use", host)
	requestno := fmt.Sprintf("%s%s%s", c.mchid, time.Now().Format("20060102150405"), RandString(8))
	b, err := c.DoPost(rawurl, map[string]interface{}{
		"coupon_code":    couponCode,
		"use_request_no": requestno,
		"appid":          c.AppID,
		"use_time":       time.Now().Format("2006-01-02T15:04:05Z07:00"),
	})
	if err != nil {
		return err
	}
	resp := struct {
		StockID          string `json:"stock_id"`
		OpenID           string `json:"openid"`
		WechatpayUseTime string `json:"wechatpay_use_time"`
		Code             string `json:"code"`
		Message          string `json:"message"`
	}{}
	if err := jsoniter.ConfigCompatibleWithStandardLibrary.Unmarshal(b, &resp); err != nil {
		return err
	}

	if resp.StockID == "" {
		return errors.New(resp.Message)
	}
	return nil
}
